unit BtnAnimado;

interface

uses Windows, Controls, Messages, Classes, Graphics, Types,
  Forms, SysUtils;

const
  TmpEfeito = 30;
  DifBordaVisivel = 4; // Diferena entre a borda visvel antes e depois de o mouse passar por cima.
  DistRetangFoco = 6; // Distncia entre a borda e o retngulo que determina o foco
  CurvaturaRetangulo = 12;

type
  TEstadoBotao = (ebNormal, ebIluminado, ebPressionado);

  TTipoEfeito = (teTransicao, teClique);
  TTipoEfeitoTransicao = (teNormalizar, teIluminar, tePressionar, teLevantar);

  TThrEfeitosBtnAnimado = class(TThread)
  private
    FBtnAnimado: TObject;
    FEstadoAnterior: TEstadoBotao;
    FTipoEfeito: TTipoEfeito;
    FTipoEfeitoTransicao: TTipoEfeitoTransicao;

    tmpCor: TColor;
    tmpBrush: TBrush;
    tmpRetangulo: TRect;
    tmpCaneta: TPen;

    vChamaOnAumenta, vChamaOnDiminui: Boolean;

    procedure SetCorCanetaBtn;
    procedure SetBrushBtn;
    procedure SetPenBtn;
    procedure FazRetangulo;
    procedure PegaAreaClienteDifBorda;
    procedure ChamaOnAumenta;
    procedure ChamaOnDiminui;

  public
    procedure FazTransicao(BtnAnimado: TObject; EstadoAnterior: TEstadoBotao;
    TipoEfeitoTransicao: TTipoEfeitoTransicao);
    procedure EfeitoClique(BtnAnimado: TObject);
  protected
    procedure DoTerminate; override;
    procedure Execute; override; // No chamar esta rotina externamente
  end;


  TBtnAnimado = class(TCustomControl)
  private
    EstadoBotao: TEstadoBotao;
    FCorBorda, FCorBordaIluminada: TColor;
    FTxtRotulo: String;
    ThrEfeitoTransicao: TThrEfeitosBtnAnimado;

    procedure FinalizaThrEfeito;
    procedure SetCorBorda(Value: TColor);
    procedure SetTxtRotulo(Value: String);
    procedure TrocaEstadoFoco;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure WMLButtonDown(var Message: TMessage); message WM_LButtonDown;
    procedure WMKeyDown(var Message: TWmKeyDown); message WM_KeyDown;
    procedure CMEnter(var Message: TMessage); message CM_ENTER;
    procedure CMExit(var Message: TMessage); message CM_EXIT;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  protected
    FValorEfeitoAcumulado:Integer;
    FOnAumenta, FOnDiminui: TNotifyEvent;
    procedure PintaTexto;
    procedure Paint; override;
    function AreaClienteDifBorda: TRect;
    function SeFocado: Boolean;
  published
    property TxtRotulo: String read FTxtRotulo write SetTxtRotulo;
    property CorBorda: TColor read FCorBorda write SetCorBorda;
    property CorBordaIluminada: TColor read FCorBordaIluminada write FCorBordaIluminada;
    property OnAumenta: TNotifyEvent read FOnAumenta write FOnAumenta;
    property OnDiminui: TNotifyEvent read FOnDiminui write FOnDiminui;

    property Enabled;
    property Font;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnClick;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Componentes do usurio', [TBtnAnimado]);
end;

{ TThrEfeitosBtnAnimado }

procedure TThrEfeitosBtnAnimado.ChamaOnAumenta;
begin
  (FBtnAnimado as TBtnAnimado).FOnAumenta(FBtnAnimado)
end;

procedure TThrEfeitosBtnAnimado.ChamaOnDiminui;
begin
  (FBtnAnimado as TBtnAnimado).FOnDiminui(FBtnAnimado)
end;

procedure TThrEfeitosBtnAnimado.DoTerminate;
begin
  inherited;
  if vChamaOnAumenta then
    Synchronize(ChamaOnAumenta);
  if vChamaOnDiminui then
    Synchronize(ChamaOnDiminui);
end;

procedure TThrEfeitosBtnAnimado.EfeitoClique(BtnAnimado: TObject);
begin
  FBtnAnimado := BtnAnimado;
  FTipoEfeito := teClique;
  Resume;
end;

procedure TThrEfeitosBtnAnimado.Execute;
var
  I, VlrAcumulado: Integer;
  R, R2: TRect;
  VlrIncrementa: Real;
begin
  inherited;
  vChamaOnAumenta := False;
  vChamaOnDiminui := False;
  VlrAcumulado := 0;
  case FTipoEfeito of
    teTransicao: begin
      case FTipoEfeitoTransicao of
        teIluminar: begin
          case FEstadoAnterior of
            ebNormal: begin
              with FBtnAnimado as TBtnAnimado do
              begin
                Synchronize(PegaAreaClienteDifBorda);
                R2 := tmpRetangulo;
                VlrIncrementa := DifBordaVisivel / TmpEfeito;
                tmpCor := FCorBordaIluminada;
                Synchronize(SetCorCanetaBtn);
                for I := FValorEfeitoAcumulado to TmpEfeito do
                begin
                  if Terminated then
                  begin
                    FValorEfeitoAcumulado := I;
                    Exit;
                  end;
                  R := R2;
                  R.Left := R.Left - Round(VlrIncrementa * I);
                  R.Top := R.Top - Round(VlrIncrementa * I);
                  R.Right := R.Right + Round(VlrIncrementa * I);
                  R.Bottom := R.Bottom + Round(VlrIncrementa * I);
                  tmpRetangulo := R;
                  Synchronize(FazRetangulo);

                  Sleep(1);
                  VlrAcumulado := I;
                end;
                if not Assigned(tmpBrush) then
                  tmpBrush := TBrush.Create;
                tmpBrush.Style := bsSolid;
                tmpBrush.Color := clInfoBk;
                Synchronize(SetBrushBtn);
                tmpRetangulo := GetClientRect;
                Synchronize(FazRetangulo);
                if Assigned(FOnAumenta) then
                  vChamaOnAumenta := True;
              end;
            end;
          end;
        end;
        teNormalizar: begin
          case FEstadoAnterior of
            ebIluminado: begin
              with FBtnAnimado as TBtnAnimado do
              begin
                Refresh;
                if not Assigned(tmpBrush) then
                  tmpBrush := TBrush.Create;
                tmpBrush.Style := bsClear;
                Synchronize(SetBrushBtn);
                Synchronize(PegaAreaClienteDifBorda);
                Synchronize(PintaTexto);
                R2 := tmpRetangulo;
                VlrIncrementa := DifBordaVisivel / TmpEfeito;
                for I := FValorEfeitoAcumulado downto 0 do
                begin
                  if Terminated then
                  begin
                    FValorEfeitoAcumulado := I;
                    Exit;
                  end;

                  tmpCor:=GetParentForm(FBtnAnimado as TBtnAnimado).Color;
                  Synchronize(SetCorCanetaBtn);

                  R := R2;
                  R.Left := R.Left - Round(VlrIncrementa * (I + 1));
                  R.Top := R.Top - Round(VlrIncrementa * (I + 1));
                  R.Right := R.Right + Round(VlrIncrementa * (I + 1));
                  R.Bottom := R.Bottom + Round(VlrIncrementa * (I + 1));
                  tmpRetangulo := R;
                  Synchronize(FazRetangulo);

                  tmpCor := FCorBorda;
                  Synchronize(SetCorCanetaBtn);

                  R := R2;
                  R.Left := R.Left - Round(VlrIncrementa * I);
                  R.Top := R.Top - Round(VlrIncrementa * I);
                  R.Right := R.Right + Round(VlrIncrementa * I);
                  R.Bottom := R.Bottom + Round(VlrIncrementa * I);
                  tmpRetangulo := R;
                  Synchronize(FazRetangulo);

                  Sleep(1);
                  VlrAcumulado := I;
                end;
                if Assigned(FOnDiminui) then
                  vChamaOnDiminui := True;
              end;
            end;
          end;
        end;
      end;
    end;
    teClique:begin
      with FBtnAnimado as TBtnAnimado do
      begin
        tmpRetangulo := GetClientRect;
        tmpRetangulo.Left := tmpRetangulo.Left + Canvas.Pen.Width + DifBordaVisivel + 5;
        tmpRetangulo.Top := tmpRetangulo.Top + Canvas.Pen.Width + DifBordaVisivel + 5;
        tmpRetangulo.Right := tmpRetangulo.Right - Canvas.Pen.Width - DifBordaVisivel - 5;
        tmpRetangulo.Bottom := tmpRetangulo.Bottom - Canvas.Pen.Width - DifBordaVisivel - 5;
        tmpBrush := TBrush.Create;
        tmpBrush.Style := bsClear;
        Synchronize(SetBrushBtn);
        tmpCaneta := TPen.Create;
        tmpCaneta.Width := 2;
        tmpCaneta.Mode := pmNot;
        Synchronize(SetPenBtn);
        Synchronize(FazRetangulo);
        Sleep(30);
        Synchronize(FazRetangulo);
        Sleep(50);
        Synchronize(FazRetangulo);
        Sleep(30);
        Synchronize(FazRetangulo);
        tmpCaneta.Mode := pmCopy;
        tmpCaneta.Width := 1;
        Synchronize(SetPenBtn);
        FreeAndNil(tmpCaneta);
        FreeAndNil(tmpBrush);
      end;
    end;
  end;
  if FTipoEfeito <> teClique then
    (FBtnAnimado as TBtnAnimado).FValorEfeitoAcumulado := VlrAcumulado;
  Terminate;
  (FBtnAnimado as TBtnAnimado).Refresh;
end;

procedure TThrEfeitosBtnAnimado.FazRetangulo;
begin
//  (FBtnAnimado as TBtnAnimado).Canvas.Rectangle(tmpRetangulo);
  (FBtnAnimado as TBtnAnimado).Canvas.RoundRect(tmpRetangulo.Left, tmpRetangulo.Top, tmpRetangulo.Right, tmpRetangulo.Bottom, CurvaturaRetangulo, CurvaturaRetangulo);
end;

procedure TThrEfeitosBtnAnimado.FazTransicao(BtnAnimado: TObject;
  EstadoAnterior: TEstadoBotao; TipoEfeitoTransicao: TTipoEfeitoTransicao);
begin
  FBtnAnimado := BtnAnimado;
  FEstadoAnterior := EstadoAnterior;
  FTipoEfeito := teTransicao;
  FTipoEfeitoTransicao := TipoEfeitoTransicao;
  Resume;
end;

procedure TThrEfeitosBtnAnimado.PegaAreaClienteDifBorda;
begin
  tmpRetangulo := (FBtnAnimado as TBtnAnimado).AreaClienteDifBorda;
end;

procedure TThrEfeitosBtnAnimado.SetBrushBtn;
begin
  (FBtnAnimado as TBtnAnimado).Canvas.Brush := tmpBrush;
end;

procedure TThrEfeitosBtnAnimado.SetCorCanetaBtn;
begin
  (FBtnAnimado as TBtnAnimado).Canvas.Pen.Color := tmpCor;
end;

procedure TThrEfeitosBtnAnimado.SetPenBtn;
begin
  (FBtnAnimado as TBtnAnimado).Canvas.Pen := tmpCaneta;
end;

{ TBtnAnimado }

function TBtnAnimado.AreaClienteDifBorda: TRect;
var
  R: TRect;
begin
  R := GetClientRect;
  R.Left := R.Left + DifBordaVisivel;
  R.Top := R.Top + DifBordaVisivel;
  R.Right := R.Right - DifBordaVisivel;
  R.Bottom := R.Bottom - DifBordaVisivel;
  Result := R;
end;

procedure TBtnAnimado.CMEnter(var Message: TMessage);
begin
  TrocaEstadoFoco;
end;

procedure TBtnAnimado.CMExit(var Message: TMessage);
begin
  TrocaEstadoFoco;
end;

procedure TBtnAnimado.CMMouseEnter(var Message: TMessage);
var
  EstadoAnterior: TEstadoBotao;
begin
  if EstadoBotao <> ebIluminado then
  begin
    FinalizaThrEfeito;
    EstadoAnterior := EstadoBotao;
    EstadoBotao := ebIluminado;
    ThrEfeitoTransicao.FazTransicao(Self, EstadoAnterior, teIluminar);
  end;
end;

procedure TBtnAnimado.CMMouseLeave(var Message: TMessage);
var
  EstadoAnterior: TEstadoBotao;
begin
  if EstadoBotao <> ebNormal then
  begin
    FinalizaThrEfeito;
    EstadoAnterior := EstadoBotao;
    EstadoBotao := ebNormal;
    ThrEfeitoTransicao.FazTransicao(Self, EstadoAnterior, teNormalizar);
  end;
end;

constructor TBtnAnimado.Create(AOwner: TComponent);
begin
  inherited;
  Width := 83;
  Height := 33;
  EstadoBotao := ebNormal;
  FCorBordaIluminada := clRed;
  FTxtRotulo := Name;
  Canvas.Font.Name := 'Arial';
  TabStop := True;
end;

destructor TBtnAnimado.Destroy;
begin
  if Assigned(ThrEfeitoTransicao) then
    if not ThrEfeitoTransicao.Terminated then
    begin
      ThrEfeitoTransicao.Terminate;
      ThrEfeitoTransicao.WaitFor;
    end;
  FreeAndNil(ThrEfeitoTransicao);
  inherited;
end;

procedure TBtnAnimado.FinalizaThrEfeito;
begin
  if Assigned(ThrEfeitoTransicao) then
  begin
    ThrEfeitoTransicao.Terminate;
    ThrEfeitoTransicao.WaitFor;
    FreeAndNil(ThrEfeitoTransicao);
  end;
  ThrEfeitoTransicao := TThrEfeitosBtnAnimado.Create(True);
end;

procedure TBtnAnimado.Paint;
var
  R: TRect;
begin
  inherited;

  if Assigned(ThrEfeitoTransicao) then
    if not ThrEfeitoTransicao.Terminated then
      Exit;

  R := ClientRect;
  case EstadoBotao of
    ebNormal:
    begin
      Canvas.Brush.Style := bsClear;
      Canvas.Pen.Color := FCorBorda;
      Canvas.RoundRect(AreaClienteDifBorda.Left, AreaClienteDifBorda.Top,
        AreaClienteDifBorda.Right, AreaClienteDifBorda.Bottom, CurvaturaRetangulo, CurvaturaRetangulo);
      PintaTexto;
    end;
    ebIluminado:
    begin
      Canvas.Brush.Style := bsSolid;
      Canvas.Pen.Color := FCorBordaIluminada;
      Canvas.RoundRect(GetClientRect.Left, GetClientRect.Top,
        GetClientRect.Right, GetClientRect.Bottom, CurvaturaRetangulo, CurvaturaRetangulo);
      PintaTexto;
    end;
  end;
  if Focused then
    TrocaEstadoFoco;
end;

procedure TBtnAnimado.PintaTexto;
begin
  Canvas.Font := Font;
  if EstadoBotao = ebNormal then
    Canvas.Font.Style := Canvas.Font.Style - [fsBold]
  else
  if EstadoBotao = ebIluminado then
    Canvas.Font.Style := Canvas.Font.Style + [fsBold];
  Canvas.Font.Color := clSilver;
  Canvas.TextOut((Width div 2) - (Canvas.TextWidth(FTxtRotulo) div 2) + 2,
                 (Height div 2) - (Canvas.TextHeight(FTxtRotulo) div 2) + 2,
                 FTxtRotulo);
  Canvas.Font.Color := clBlue;
  Canvas.TextOut((Width div 2) - (Canvas.TextWidth(FTxtRotulo) div 2),
                 (Height div 2) - (Canvas.TextHeight(FTxtRotulo) div 2),
                 FTxtRotulo);
end;

function TBtnAnimado.SeFocado: Boolean;
begin
  Result := Focused;
end;

procedure TBtnAnimado.SetCorBorda(Value: TColor);
begin
  FCorBorda := Value;
  Refresh;
end;

procedure TBtnAnimado.SetTxtRotulo(Value: String);
begin
  FTxtRotulo := Value;
  Refresh;
end;

procedure TBtnAnimado.TrocaEstadoFoco;
var
  R: TRect;
begin
  R := ClientRect;
  R.Left := R.Left + DistRetangFoco;
  R.Top := R.Top + DistRetangFoco;
  R.Right := R.Right - DistRetangFoco;
  R.Bottom := R.Bottom - DistRetangFoco;
  Canvas.Pen.Style := psDot;
  Canvas.Pen.Mode := pmCopy;
  if Focused then
    Canvas.Pen.Color := clBlue
  else
    Canvas.Pen.Color := GetParentForm(Self).Color;
  Canvas.Brush.Style := bsClear;
  Canvas.RoundRect(R.Left, R.Top, R.Right, R.Bottom, CurvaturaRetangulo, CurvaturaRetangulo);
//  Canvas.DrawFocusRect(R);
//  Canvas.Brush.Style := bsSolid;
  Canvas.Pen.Mode := pmCopy;
  Canvas.Pen.Style := psSolid;
end;

procedure TBtnAnimado.WMKeyDown(var Message: TWmKeyDown);
begin
  inherited;
  if Message.CharCode = 13 then
    if Assigned(OnClick) then
      OnClick(Self);
end;

procedure TBtnAnimado.WMLButtonDown(var Message: TMessage);
begin
  inherited;
  if EstadoBotao = ebIluminado then
  begin
    FinalizaThrEfeito;
    ThrEfeitoTransicao.EfeitoClique(Self);
  end;
  SetFocus;
end;

end.
